/* -*- Mode: C++ -*-
 * $Id: AssistOffensiveIntention.cc,v 1.8 2001/07/29 11:41:06 mikbr212 Exp $
 *
 * IDENTIFICATION
 *
 * $Revision: 1.8 $
 * Created by: Mikael Brnnstrm, Andreas Bckert, Andreas Sigfridsson
 * Last modified by $Author: mikbr212 $ $Date: 2001/07/29 11:41:06 $
 *
 *
 * HISTORY
 *
 * $Log: AssistOffensiveIntention.cc,v $
 * Revision 1.8  2001/07/29 11:41:06  mikbr212
 * Not that agressive aymore
 *
 * Revision 1.7  2001/07/29 09:43:30  mikbr212
 * Adjusted some things
 *
 * Revision 1.6  2001/07/26 12:19:33  mikbr212
 * Changed passtree
 *
 * Revision 1.5  2001/07/06 20:48:27  mikbr212
 * Added some reasonable utilities
 *
 * Revision 1.4  2001/07/06 14:39:20  mikbr212
 * fixed utility
 *
 * Revision 1.3  2001/07/06 08:30:53  mikbr212
 * Made the intention work (but not yet in the new design)
 *
 * Revision 1.2  2001/07/03 15:24:18  andbo928
 * Added utility methods.
 *
 * Revision 1.1  2001/06/19 21:35:10  mikbr212
 * Created stubs
 *
 * Revision 1.1  2001/06/19 14:11:43  mikbr212
 * Created
 *
 */

/**
 * @file AssistOffensiveIntention.cc
 * Implementation file for the class AssistOffensiveIntention.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.8 $
 * $Author: mikbr212 $
 * $Date: 2001/07/29 11:41:06 $
 */


#include "AssistOffensiveIntention.h"

#include "MoveToPosition.h"
#include "Role.h"

RS_BEGIN_NAMESPACE

/** Constructs an intention. 
 */
AssistOffensiveIntention::AssistOffensiveIntention()
  : Intention()
{
}

/** Destroys the intention. 
 */
AssistOffensiveIntention::~AssistOffensiveIntention() {
}

/* Chooses a skill. It works like a mini-decision maker.
 * It will only use the move to position skill. The position
 * given in getPosition will be used and the utility from
 * getNeedOfAssistance will be used to set the speed.
 */
Skill* AssistOffensiveIntention::chooseSkill() {
  static MoveToPosition moveSkill;
  moveSkill.setPosition(getPosition());
  Utility util(getNeedOfAssistance(worldStateView->getWorldState()));
  if(util >= uNormal)
    moveSkill.setSpeed(MoveToPosition::msNormal);
  else
    moveSkill.setSpeed(MoveToPosition::msKeepStamina);

  return &moveSkill;
}

/** Models a player. 
 * The player will run upwards, until it is very far up.
 */
void AssistOffensiveIntention::model(WorldState& ws) {
  const float maxX(45);
  AgentState agent(ws.getAgent());
  if(agent.getPosition().getX() > maxX) {
    ws.doNull();
  } else {
    agent.setDirection(0); // up (cheat a little)
    ws.setAgent(agent);
    Utility util(getNeedOfAssistance(ws));
    if(util >= uNormal)
      ws.doDash(50);
    else
      ws.doDash(30);
  }
}

/** Checks if it is a special offensive situation, ie when
 * it is our kick off and the agent is an offensive player.
 */
bool AssistOffensiveIntention::isSpecialSituation() {
  static Role role;
  if(role.getRoleType() != Role::rtOffensive)
    return false;
  
  if(statisticsView->getCyclesSinceOurKickoff() < 20)
    return true;
  return false;
}

/** Computes a position that is slightly above the ball. If a position
 * was computed last cycle then the direction to that position 
 * will get a little "bonus", so that the agent doesn't need to
 * make that turn it otherwise would need.
 */
TVector AssistOffensiveIntention::getPosition() {
  const float distanceAhead(1); // distance ahead of ball
  const float coneAngle(60.0); // angle of the cone of valid positions
  const float dy(5.0); // distance between positions (on the line)

  static TVector lastPosition;
  static int time(0);

  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());
  const TVector agentPosition(agent.getPosition());

  PassTree pt(passTreeView->getAgentPassTree());
  float bestValue(0);
  TVector bestPosition;

  float value;
  TVector position;

  // Compute positions on a line slightly ahead of the ball
  // Then evalueate them
  float x(ball.getPosition().getX() - agent.getPosition().getX()
	  + distanceAhead);

  if(isSpecialSituation()) 
    x += 10;
    
  if(x <= 0) 
    return agentPosition;
  
  float maxY(x/std::tan(coneAngle/2*M_PI/180.0));
  for(float y = -maxY; y < maxY; y += dy) {
    position = agentPosition + TVector(x,y);
    value = pt.evaluatePosition(position);
    if(value > bestValue) {
      bestValue = value;
      bestPosition = position;
    }
  }

  // Check last position (if valid)
  // It is better to continue to run in the same dir
  int curTime = worldStateView->getWorldState().getGameTime();
  if(curTime - time <= 2) {
    // Compute a new position in the same direction
    TVector delta(lastPosition - agentPosition);
    float newX(ball.getPosition().getX() - agentPosition.getX() 
	       + distanceAhead);
    // scale the delta-vector so that the x-coord = newX
    delta *= newX / delta.getX();
    position = agentPosition + delta;
    value = pt.evaluatePosition(position);
    if(value + 0.05 > bestValue) {
      bestValue = value;
      bestPosition = position;
    }
  }
  
  if(bestValue > 0.01) {
    time = curTime;
    lastPosition = bestPosition;
    return bestPosition;
  }

  return agentPosition; 
}

/** Computes and returns a utility, that describes the need of assistance
 * above the ball. 
 * The value is cached and is dependent on the positions of the players
 * only. Therefore the gameTime in the worldstate is used to check if 
 * a new value has to be calculated. It works for both teams.
 */
Intention::Utility AssistOffensiveIntention::getNeedOfAssistance(const WorldState& ws) {
  static Utility tmUtility(uSuck);
  static Utility opUtility(uSuck);
  static int time(0);

  int curTime(ws.getGameTime());
  if(time == curTime) 
    return (ws.getTeam() == T_OUR_TEAM ? tmUtility : opUtility);

  const float ballMargin(3.0);
  float referenceX(ws.getBall().getPosition().getX());
  if(referenceX > 20)
    referenceX = 20;
  float ballPosXMin(referenceX - ballMargin);
  float ballPosXMax(referenceX + ballMargin);
  std::vector<PlayerState>::const_iterator it;

  // *** Check teammates ***
  int tmDefCount(0);
  int tmMidCount(0);
  int tmOffCount(0);
  const std::vector<PlayerState> tm(ws.getTeammates());
  for(it=tm.begin(); it!=tm.end(); it++) {
    if(it->getPosition().getX() > ballPosXMax)
      tmOffCount++;
    else if(it->getPosition().getX() > ballPosXMin)
      tmMidCount++;
    else
      tmDefCount++;
  }

  // *** Check opponents ***
  int opDefCount(0);
  int opMidCount(0);
  int opOffCount(0);
  const std::vector<PlayerState> op(ws.getOpponents());
  for(it=op.begin(); it!=op.end(); it++) {
    if(it->getPosition().getX() < ballPosXMin)
      opOffCount++;
    else if(it->getPosition().getX() < ballPosXMax)
      opMidCount++;
    else
      opDefCount++;
  }

  // If this was their team, then swap op and tm values
  if(ws.getTeam() != T_OUR_TEAM) {
    int tmp;
    tmp = tmDefCount;
    tmDefCount = opDefCount;
    opDefCount = tmp;

    tmp = tmMidCount;
    tmMidCount = opMidCount;
    opMidCount = tmp;

    tmp = tmOffCount;
    tmOffCount = opOffCount;
    opOffCount = tmp;
  }

  // *** Interpret the values ***
  
  // Mid-players can be both offensive and defensive
  tmOffCount += tmMidCount;
  tmDefCount += tmMidCount;
  opOffCount += opMidCount;
  opDefCount += opMidCount;

  const int many(8);
  const int few(3);

  
  // ** teammates **
  if(opDefCount >= many) { 
    // *** Many opponents ***
    if(tmOffCount >= many) { // Many teammates
      tmUtility = uBad; 
    } else if(tmOffCount <= few) { // Few teammates
      tmUtility = uNormal; 
    } else { // Not few, not many teammates
      tmUtility = uGood; 
    }
  } else if(opDefCount <= few) { 
    // *** Few opponents ***
    if(tmOffCount >= many) { // Many teammates
      tmUtility = uVeryBad; 
    } else if(tmOffCount <= few) { // Few teammates
      tmUtility = uVeryGood; 
    } else { // Not few, not many teammates
      tmUtility = uNormal; 
    }
  } else { 
    // *** Not few, not many opponents ***
    if(tmOffCount >= many) { // Many teammates
      tmUtility = uBad; 
    } else if(tmOffCount <= few) { // Few teammates
      tmUtility = uGood; 
    } else { // Not few, not many teammates
      tmUtility = uNormal; 
    }
  }

  // ** opponents **
  if(tmDefCount >= many) { 
    // *** Many opponents ***
    if(opOffCount >= many) { // Many teammates
      opUtility = uBad; 
    } else if(opOffCount <= few) { // Few teammates
      opUtility = uNormal; 
    } else { // Not few, not many teammates
      opUtility = uGood; 
    }
  } else if(tmDefCount <= few) { 
    // *** Few opponents ***
    if(opOffCount >= many) { // Many teammates
      opUtility = uVeryBad; 
    } else if(opOffCount <= few) { // Few teammates
      opUtility = uVeryGood; 
    } else { // Not few, not many teammates
      opUtility = uNormal; 
    }
  } else { 
    // *** Not few, not many opponents ***
    if(opOffCount >= many) { // Many teammates
      opUtility = uBad; 
    } else if(opOffCount <= few) { // Few teammates
      opUtility = uGood; 
    } else { // Not few, not many teammates
      opUtility = uNormal; 
    }
  }
  
  time = curTime;
  return (ws.getTeam() == T_OUR_TEAM ? tmUtility : opUtility);
}

/** Checks if the utility of this intention.
 */
Intention::Utility AssistOffensiveIntention::modelUtility(const WorldState& ws) {
  const BallState& ball(ws.getBall());
  const AgentState& agent(ws.getAgent());

  float xDist(ball.getPosition().getX() - agent.getPosition().getX());

  // Check if the player already is taking part in the offensive
  if(xDist < 0) // agent above ball
    return uSuck; // yepp, then get passable (etc) instead
  
  return commonUtility(ws);
}

/** The things that are shared by modelUtility and utility are here.
 */
Intention::Utility 
AssistOffensiveIntention::commonUtility(const WorldState& ws) {
  const BallState& ball(ws.getBall());
  const AgentState& agent(ws.getAgent());

  float xDist(ball.getPosition().getX() - agent.getPosition().getX());
  Utility util(getNeedOfAssistance(ws));

  // Punish distant players
  if(xDist > 22.0)
    util = Utility(max(int(uVeryBad), int(util-2)));
  else if(xDist > 13.0)
    util = Utility(max(int(uVeryBad), int(util-1)));

  return util;
}

/** Checks if it is an offensive situation. Then commonUtility
 * is called.
 */
Intention::Utility AssistOffensiveIntention::utility() {
  const WorldState& ws(worldStateView->getWorldState());
  const BallState& ball(ws.getBall());
  const AgentState& agent(ws.getAgent());

  // Check if we (will) controll the ball (ie if it is an offensive situation)
  if(statisticsView->getOurInterceptionTime() - 2 > 
     statisticsView->getTheirInterceptionTime() ||
     gameView->isTheirBall())
    return uSuck; // nope

  float xDist(ball.getPosition().getX() - agent.getPosition().getX());
  
  if(isSpecialSituation()) {
    if(agent.getPosition().getX() > ws.getOffsideLine())
      return uSuck;
  } else {
    // Check if the player already is taking part in the offensive
    if(xDist < 0) // agent above ball
      return uSuck; // yepp, then get passable (etc) instead
  }

  return commonUtility(ws);
}

/** Returns the persistence of this intention.
 * This intention is very persistent.
 * @return a float between zero (not persistent) and one (very persistent)
 */
float AssistOffensiveIntention::getPersistence() const {
  return 0.8; // Very persistent
}

/** Returns the ID that identifies this (type of) intention.
 */
IntentionId AssistOffensiveIntention::getID() const {
  return ASSIST_OFFENSIVE_INTENTION_ID;
}

RS_END_NAMESPACE

/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "AssistOffensiveIntention.icc"
#endif







