/* -*- Mode: C++ -*-
 * $Id: CatchBall.cc,v 1.13 2001/07/31 17:09:43 andbo928 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: CatchBall.cc
 * Unitname: Framework/skill
 * $Revision: 1.13 $
 * Created by: Bjrn, Ola och Jenny
 * Last modified by $Author: andbo928 $ $Date: 2001/07/31 17:09:43 $
 *
 *
 * HISTORY
 *
 *
 */

/**
 * @file CatchBall.cc
 * This is a definition file for the skill CatchBall.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.13 $
 * $Author: andbo928 $
 * $Date: 2001/07/31 17:09:43 $
 */

#include "CatchBall.h"
#include "view_constants.h"
#include "CatchCommand.h"
#include "TurnCommand.h"
#include "ParameterView.h"
#include "InPenaltyArea.h"

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---
CatchBall::CatchBall()
  : Skill(),
    parameterView(NULL),
    gameView(NULL),
    worldStateView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);  
  getView(gameView, GAME_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
}

// --- DESTRUCTOR ---
CatchBall::~CatchBall()
{
    viewManager->releaseView(WORLD_STATE_VIEW_ID);
    viewManager->releaseView(GAME_VIEW_ID);
    viewManager->releaseView(PARAMETER_VIEW_ID);
}

// --- METHODS ---
Plan CatchBall::generatePlan()
{
  // Check out two states.
  // A) "Catch this cycle"
  // B) "Turn towards ball and catch"
  // C) "4 different dashes."

  // A)
  WorldState thisCycle = worldStateView->getWorldState();
  double catchNowFactor = evaluate(thisCycle);
  //  cout << "now : " << catchNowFactor << endl;

  // B)
  WorldState nextCycle = thisCycle;
  nextCycle.doTurnTowardsBall();
  nextCycle.doTurnNeckTowardsBall();
  Step turnStep = nextCycle.getStep();
  nextCycle.step();

  double catchNextFactor = evaluate(nextCycle);
  //  cout << "next: " << catchNextFactor << endl;

  // Choose the best alternative.

  planTime = worldStateView->getWorldState().getGameTime();
  if(catchNextFactor > catchNowFactor
     && catchNowFactor < 0.3)
    {
      //      cout << "turn" << endl;
      planValue = catchNextFactor;
      thePlan = Plan(turnStep);
    }
  else
    {
      //      cout << "catch" << endl;
      planValue = catchNowFactor;
      lastCatchTime = worldStateView->getWorldState().getGameTime();
      TVector catchVec = calculateCatchVector(thisCycle);
      CatchCommand cmd(catchVec.getAngle());
      thePlan = Plan(Step(&cmd));
    }

  return thePlan;
}

FuzzyBool CatchBall::persistent()
{
  return FALSE;
}

TVector CatchBall::calculateCatchVector(const WorldState& worldState)
{
  TVector ballPos = worldState.getBall().getPosition();
//    TVector ballSpeed = worldState.getBall().getSpeed();
  TVector agentPos = worldState.getAgent().getPosition();
//    TVector agentSpeed = worldState.getAgent().getSpeed();

  double agentDir = worldState.getAgent().getDirection();

  TVector AtoB = ballPos - agentPos;
  AtoB.rotate(-agentDir);

  return AtoB;
}

double CatchBall::evaluate(const WorldState& worldState)
{
  TVector catchVec = calculateCatchVector(worldState);
  double factorL;

  factorL = abs(catchVec.getLength() / (parameterView->SP_catchable_area_l));
  if (factorL > 0.5)
    factorL = 1 - factorL;
  factorL *= 2;
  factorL = max(0.0, factorL );
  if(catchVec.getAngle() < -90.0 || catchVec.getAngle() > 90.0)
    factorL = 0.0;

  return factorL;
}

bool CatchBall::isBanned()
{
  // Check that we aren't in the catch_ban cycles
  if(worldStateView->getWorldState().getGameTime()
     - lastCatchTime < parameterView->SP_catch_ban_cycle)
  {
    return true;
  }
  return false;
}

FuzzyBool CatchBall::applicable()
{
  static InPenaltyArea inPenaltyArea;

  // Need to be a goalie! duh!
  if(!parameterView->CP_goalie)
    return FALSE;

  // Only catch in penaltyarea.

  if(inPenaltyArea() < FuzzyBool(0.75))
    return FALSE;

  // We need to be pretty sure where the ball is.
  if(worldStateView->getWorldState().getBall().getC() < 0.5)
    return FALSE;

  // Check that we aren't in the catch_ban cycles
  if(isBanned())
  {
    return FALSE;
  }

  // Foo!

  WorldState thisCycle = worldStateView->getWorldState();
  double catchNowFactor = evaluate(thisCycle);

  WorldState nextCycle = thisCycle;
  nextCycle.doTurnNeckTowardsBall();
  Step turnStep = nextCycle.getStep();
  nextCycle.step();
  double catchNextFactor = evaluate(nextCycle);

  return max(catchNowFactor, catchNextFactor);

}

RS_END_NAMESPACE
