/* -*- Mode: C++ -*-
 * $Id: Dribble.cc,v 1.13 2001/07/31 14:26:42 andbo928 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallToSpeedVector.cc
 * Unitname: Framework/skill
 * $Revision: 1.13 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: andbo928 $ $Date: 2001/07/31 14:26:42 $
 *
 *
 * HISTORY
 *
 * $Log: Dribble.cc,v $
 * Revision 1.13  2001/07/31 14:26:42  andbo928
 * Added setSafeCycles.
 *
 * Revision 1.12  2001/07/30 16:49:06  andbo928
 * Added setSpeed method.
 *
 * Revision 1.11  2001/07/30 12:32:05  andbo928
 * Fixed some spelling errors.
 *
 * Revision 1.10  2001/07/30 11:00:50  andbo928
 * Implemented.
 *
 * Revision 1.9  2001/07/29 13:11:55  mikbr212
 * fixed the basics
 *
 * Revision 1.8  2001/07/29 12:48:37  mikbr212
 * New skill
 *
 *
 */

/**
 * @file Dribble.cc
 * Implementation file for the skill class Dribble.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.13 $
 * $Author: andbo928 $
 * $Date: 2001/07/31 14:26:42 $
 */

#include "Dribble.h"
#include "TAngle.h"
#include "AngleThingy.h"
#include "KickBallToSpeedVector.h"

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

Dribble::Dribble()
  : Skill(),
    parameterView(NULL),
    worldStateView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  dribbleSpeed = 0.4;
  maxSteps = 12;
}

// --- DESTRUCTOR ---

Dribble::~Dribble()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}

// --- METHODS ---

/** Checks if an opponent is close to the agent.
 */
bool Dribble::isOpponentClose(float nearDist) {
  const WorldState& ws(worldStateView->getWorldState());
  const std::vector<PlayerState>& op(ws.getOpponents());
  const TVector agentPos(ws.getAgent().getPosition());
  
  std::vector<PlayerState>::const_iterator it;
  for(it = op.begin(); it != op.end(); it++) {
    if((it->getPosition() - agentPos).getLength() < nearDist)
      return true;
  }
  return false;
}

bool Dribble::shouldLook() {
  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const TVector goal(52.5,0);

  // check if looking towards their goal is needed
  TVector toGoal(goal - agent.getPosition());
  if(fabs(TAngle(toGoal.getAngle() - agent.getDirection())) < 40) 
    return false; // nope
  
  // check if the agent has the time to look
  if(isOpponentClose(5))
    return false; // nope, they will take the ball
  
  return true;
}

void Dribble::setSpeed(double v0)
{
  if(dribbleSpeed != v0)
    planTime = -1; // Invalidate cache.
  dribbleSpeed = v0;
}

void Dribble::setSafeCycles(int cycles)
{
  if(maxSteps != cycles);
    planTime = -1; // Invalidate cache.
  maxSteps = cycles;
}

/**
 * Dribbles.
 */
Plan Dribble::generatePlan()
{
  if(planTime == worldStateView->getWorldState().getGameTime())
    return thePlan;
  planTime = worldStateView->getWorldState().getGameTime();
  unsafe = false;

  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());
  const TVector goal(52.5,0);
  AgentState newAgent(agent);
  BallState newBall(ball);

  TVector toGoal(goal - agent.getPosition());

//    if(shouldLook()) {
    
//      // Check if I need to stop the ball first
//      AgentState testAgent(agent);
//      BallState testBall(ball);
//      testAgent.step();
//      testBall.step();
    
//      if(!testAgent.isBallKickable(testBall)) {
//        // Need to stop the ball
//        newAgent.doKickToSpeed(TVector(0,0), newBall);
//        newBall.step();
//        newAgent.doTurnTowards(newBall.getPosition());
//        return Plan(newAgent.getStep());
//      }

//      newAgent.doTurnTowards(goal);
//      newBall.step();
//      newAgent.doTurnNeckTowards(newBall.getPosition());
//      return Plan(newAgent.getStep());
//    }

  // dribble
//    toGoal.setLength(0.5);
//    newAgent.doKickToSpeed(toGoal,newBall);

  // FIXME: Don't dribble out of bounds
  // FIXME: 180 deg wraparound.
  AngleThingy angles;

  AngleThingy::Span initSpan;
  initSpan.end = toGoal.getAngle() + TAngle(50.0);
  initSpan.end = toGoal.getAngle() + TAngle(50.0);
//    initSpan.start = -170;
//    initSpan.end = 170;

  angles.addSpan(initSpan);
  initSpan.end = toGoal.getAngle() + TAngle(50.0);

  angles.addSpan(initSpan);
  angles.setInitialSpeed(dribbleSpeed);

  int time;
  double bestDir = 0.0;
  for(time = 0; time < maxSteps; time++)
  {
    angles.step();

    // Find the best "safe" direction.
    if(angles.getSpans().size() == 0)
      break;

    AngleThingy::Span bestSpan = angles.getMaxSpan();
    bestDir = (bestSpan.start + bestSpan.end) / 2.0;
  }

  if(time != maxSteps)
  {
//      cout << "Warning! Unsafe dribble." << endl;
    unsafe = true;
  }
//    cout << "Dribbling dir: " << bestDir << endl;

  TVector kickVector = TVector::TPolarVector(dribbleSpeed, bestDir);
  WorldState ws = worldStateView->getWorldState();

  KickBallToSpeedVector kickBall;
  kickBall.setVector(kickVector);
  thePlan = kickBall.generatePlan();

  return thePlan;
}

/**
 * Persistent.
 * FALSE
 */
FuzzyBool Dribble::persistent()
{
  return FALSE;
}

/**
 * Returns the applicable value, true if ball is kickable.
 */
FuzzyBool Dribble::applicable()
{
  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());

  generatePlan();
  if(agent.isBallKickable(ball))
  {
    if(unsafe)
      return FuzzyBool(0.3);
    else
      return TRUE;
  }
  return FALSE;
}

RS_END_NAMESPACE

