/**
 * @file FuzzyIntercept.cc
 * Implementation of the FuzzyIntercept skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.2 $
 * $Author: andbo928 $
 * $Date: 2001/04/13 18:16:33 $
 */

#include "FuzzyIntercept.h"
#include "KickCommand.h"
#include "KickBallToSpeedVector.h"
#include "DashToSpeedVectorFacingDirection.h"

#define KICK_RANGE 30
#define SAFE_MARGIN 2
#define CB_FRIENDLY 0.2
#define CB_OPPONENT -1.0
#define CB_ADVANCE  0.5
#define CB_PRECISION 0.0

#define MAXSPEED 2.0
#define EMERGENCY true
RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

FuzzyIntercept::FuzzyIntercept(int maxStp)
  : Skill(),
    parameterView(NULL), //,
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    playerView(NULL),
    kickBall(Vector()),
    maxSteps(maxStp)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(playerView, FUZZY_PLAYER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(ballView, FUZZY_BALL_VIEW_ID);

//    mtpfd.setEmergency(EMERGENCY);
//    mtpfd.setMaxSpeed(MAXSPEED);
//    mtpfd.setInPointRange(parameterView->SP_kickable_area * 0.8);

}

// --- DESTRUCTOR ---

FuzzyIntercept::~FuzzyIntercept()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(FUZZY_BALL_VIEW_ID);
  viewManager->releaseView(FUZZY_PLAYER_VIEW_ID);
}


// --- METHODS ---

Vector FuzzyIntercept::calculateDashVector(Point p)
{

  // Do some fuzzy running towards the position.
  Point curAgentPos = agentView->getAgentPosition(NOW).getPoint();
  Vector curAgentSpeed = agentView->getSpeedVector(NOW).getVector();
  //  Vector curAgentDir = Vector(agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle(),
  //			      1);

  // Check what a dash will accomplish.
//    Point nextAgentSpeed = Point(curAgentSpeed + curAgentDir * 0.5);
//    Point nextAgentPos = Point(curAgentPos + nextAgentSpeed);
//    nextAgentSpeed *= parameterView->SP_player_decay;

  // Check if a dash is reasonably towards the position.

  Vector curAtoP = Vector(p - curAgentPos);
  //  Vector nextAtoP = Vector(lastPos - nextAgentPos);

  curAtoP.setLength(2);
  return curAtoP;
}

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

Plan FuzzyIntercept::generatePlan()
{
  int t;

  //  cout << "Intercepting!" << endl;

  Point interceptPoint = calculateInterceptPoint(&t);
  if(!useLastPos) {
    lastPos = interceptPoint;
    if(lastPos.getX() == -9999)
      return Plan();

    lastDash = calculateDashVector(lastPos);
  } 

  Vector delta = Vector(lastPos, interceptPoint);
  if(delta.getLength() > parameterView->SP_kickable_area * 0.9) {
    lastPos = interceptPoint;
    lastDash = calculateDashVector(lastPos);
  }

//    cout << "Desired pos:" << lastPos << endl;
//    cout << "BallPos: " << Point(ballView->getAbsPosition().getVector()) << endl;
//    cout << "BallSpeed: " << Point(ballView->getAbsSpeed().getVector()) << endl;

  /*  DashToSpeedVectorFacingDirection dash(lastDash);
  dash.setMaxSpeed(MAXSPEED);
  dash.setEmergency(EMERGENCY);

  Vector relBall = ballView->getRelPosition(NOW).getVector();
  //  AngleDeg dir = relBall.getAngle() - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle();

  Point curAgentPos = agentView->getAgentPosition(NOW).getPoint();
  Vector curAtoP = Vector(lastPos - curAgentPos);
  //  AngleDeg dir = curAtoP.getAngle() - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle();
  //  AngleDeg dir = relBall.getAngle();
  dash.setFaceAngle(lastDash.getAngle());

  if(dash.applicable())
    return dash.generatePlan();
  */

  Point curAgentPos = agentView->getAgentPosition(NOW).getPoint();
  Vector curAtoP = Vector(lastPos - curAgentPos);
  AngleDeg dir = curAtoP.getAngle();

  Point curBallPos = Point(ballView->getAbsPosition(NOW).getVector());
  Vector AtoB = Vector(curBallPos - curAgentPos);
  AngleDeg ballDir = AtoB.getAngle();

  // If possible, look at the ball.
  // This way we can keep the ballconfidence high.

  if(abs((dir - ballDir).getDeg()) < 80)
    {
      dir = ballDir;
    }

//    mtpfd.setPosition(lastPos);
//    mtpfd.setFaceAngle(dir);
  
//    if(mtpfd.applicable())
//      return mtpfd.generatePlan();

  return Plan();
//    // Check for the optimal turn.

//    // Just turn towards the point for now.
//    // Probably should compensate for current agent speed.

//    //  AngleDeg dir = curAtoP.getAngle() - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle();

//    if(cmd == NULL) {
//      cmd = new TurnCommand(dir.getDeg());
//    }

//    Plan plan(cmd);
  
//    delete cmd;
//    cmd = NULL;
  //  return plan;
}

Point FuzzyIntercept::calculateInterceptPoint(int *t) {

  // Calculate the ball positions.
  Point ballPos[maxSteps];
  Point desiredPos[maxSteps];
  Point curBallPos = Point(ballView->getAbsPosition(NOW).getVector());
  Vector curBallSpeed = ballView->getAbsSpeed(NOW).getVector();
  Vector ballDir = curBallSpeed;
  ballDir.setLength(1.0);

  const int safeSteps = 0;
  for(int i = 0; i < safeSteps; i++) {
    curBallSpeed = curBallSpeed * parameterView->SP_ball_decay;
    curBallPos = curBallPos + curBallSpeed;
  }

  for(int i = 0; i < maxSteps; i++) {
    curBallSpeed = curBallSpeed * parameterView->SP_ball_decay;
    curBallPos = curBallPos + curBallSpeed;
    ballPos[i] = curBallPos;
    desiredPos[i] = curBallPos;
    if(curBallSpeed.getLength() > 0.3)
      desiredPos[i] += ballDir * parameterView->SP_kickable_area * 0.5;
  }

//    for(int i = 1; i < maxSteps; i++) {
//      if(mtpfd.reachable(desiredPos[i], i)) {
//        *t = i;
//        return desiredPos[i];
//      }
//    }
  /*
  Point agentPos = agentView->getAgentPosition(NOW).getPoint();
  for(int i = 0; i < maxSteps; i++) {
    Vector AtoP = Vector(agentPos - desiredPos[i]);
    if(AtoP.getLength() < (0.5  + 1.5 * i)) {
      return desiredPos[i];
    }
  }
  */

  return Point(-9999, -9999);
}


/**
 * Persistent.
 * FALSE
 */
FuzzyBool FuzzyIntercept::persistent()
{
  FuzzyBool res = applicable();
  useLastPos = true;
 
//    // Need to know where the ball is.
//    if(!ballView->getAbsPosition().isKnown())
//      return FALSE;
//    if(ballView->getAbsPosition(NOW).getLength().getConfidence() < 0.5)
//      return FALSE;

//    // No need to intercept if we can kick the ball.
//    if(ballView->getRelPosition(NOW).getLength() < parameterView->SP_kickable_area * 0.9)
//      return FALSE;
//    return TRUE;
  return res;
}

/**
 * Returns the applicable value
 */

FuzzyBool FuzzyIntercept::applicable()
{
  
  useLastPos = false;

  // Need to know where the ball is.
  if(!ballView->getAbsPosition().isKnown())
    return FALSE;
  if(ballView->getAbsPosition(NOW).getLength().getConfidence() < 0.5)
    return FALSE;

  // No need to intercept if we can kick the ball.
  if(ballView->getRelPosition(NOW).getLength() < parameterView->SP_kickable_area * 0.9)
    return FALSE;

  int tmp;
  Point isePoint = calculateInterceptPoint(&tmp);
  if(isePoint.getX() != -9999)
    return TRUE;
  return FALSE;
}

RS_END_NAMESPACE

