/**
 * @file MoveToPosition.cc
 * Implementation of the MoveToPosition skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.24 $
 * $Author: andbo928 $
 * $Date: 2001/07/31 14:06:27 $
 */

#include "MoveToPosition.h"
#include "KickCommand.h"
#include "KickBallToSpeedVector.h"
#include "Display.h"
#include "ViewPack.h"
#include "TAngle.h"

RS_BEGIN_NAMESPACE

extern ViewPack* viewPack;

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

MoveToPosition::MoveToPosition()
  : Skill(),
    parameterView(NULL),
    worldStateView(NULL),
    lookView(NULL),
    destination(0,0,1),
    distance(1.0)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  getView(lookView, LOOK_VIEW_ID);

  lookAtBall = true;
  speed = msNormal;
}

// --- DESTRUCTOR ---

MoveToPosition::~MoveToPosition()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(LOOK_VIEW_ID);
}

// --- METHODS ---
/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

void MoveToPosition::setLookAtBall(bool enable)
{
  lookAtBall = enable;
}

void MoveToPosition::setDistance(const double r)
{
  distance = r;
}

void MoveToPosition::setPosition(const TVector& pos)
{
  destination = pos;
}

void MoveToPosition::setSpeed(const MoveSpeed _speed)
{
  speed = _speed;
}

bool MoveToPosition::goalFound(const WorldState& world)
{
  TVector delta = world.getAgent().getPosition() - destination;
  if(delta.getLength() < distance)
    return true;
  return false;
}

double MoveToPosition::goalVal(const TVector& a, const TVector& b)
{
  TVector delta = a - b;
  return delta*delta;
}

int NUM_STEPS = 16;

void MoveToPosition::staminaDash(WorldState& ws)
{
  // Heavy black magic?
  double min_stamina = (parameterView->SP_stamina_max - 
			parameterView->SP_stamina_min) * 
    (parameterView->SP_effort_inc_thr + parameterView->SP_effort_dec_thr)/2
    + parameterView->SP_stamina_min;

  switch (speed)
  {
  case msVeryFast:
    if(ws.getAgent().getStamina() > (0.9 * min_stamina + 100))
      ws.doDash(100);
    else
      ws.doTurn(0);
    break;
  case msFast:
    if(ws.getAgent().getStamina() > (min_stamina + 100))
      ws.doDash(100);
    else
      ws.doTurn(0);
    break;
  case msNormal:
    if(ws.getAgent().getStamina() > (1.1 * min_stamina + 100))
      ws.doDash(100);
    else
      ws.doTurn(0);
    break;
  case msKeepStamina:
    ws.doDash(ws.getAgent().getRecovery() * parameterView->SP_stamina_inc_max);
    break;
  case msGainStamina:
    // If we already have maximum stamina, use msKeepStamina strategy
    if(ws.getAgent().getStamina() >
       parameterView->SP_stamina_max - 1.5*parameterView->SP_stamina_inc_max)
    {
      ws.doDash(ws.getAgent().getRecovery()
		* parameterView->SP_stamina_inc_max);
    } else {
      ws.doDash(0.5 * ws.getAgent().getRecovery()
		* parameterView->SP_stamina_inc_max);
    }
    break;
  }

}

double MoveToPosition::evalDir(WorldState& ws)
{
  // Observe the reference!!

  // Just madly dashes away from a given worldstate.
  int i;
  for(i = 1; i < NUM_STEPS && !goalFound(ws); i++)
    {
      staminaDash(ws);
      ws.step();
    }
  return i * goalVal(ws.getAgent().getPosition(),
		     destination);
}

Plan MoveToPosition::generatePlan()
{
  WorldState curWorld = worldStateView->getWorldState();
  WorldState tmpWorld;
  curWorld.reduceCR(0.3, 15);

  TVector tmp = curWorld.getAgent().getPosition() - destination;
  NUM_STEPS = max(4.0, tmp.getLength() / 2);

  // Process lookview.
  Float f;
  TAngle lookDir;
  TAngle leftLimit, rightLimit;

  Step bestStep;
  double bestEval = 10000000000.0;
  Step curStep;
  double curEval;

  // First eval dash-dash.
  tmpWorld = curWorld;
  //  tmpWorld.doDash(100);
  staminaDash(tmpWorld);

  if(lookAtBall)
    {
      tmpWorld.doTurnNeckTowardsBall();
    }
  else
    {
      leftLimit = tmpWorld.getAgent().getDirection() - 85;
      rightLimit = tmpWorld.getAgent().getDirection() + 85;
      lookDir = lookView->getAngle(f, 
				   AngleDeg((double)leftLimit), 
				   AngleDeg((double)rightLimit)).getDeg();
      //    cout << "anglein: " << tmpWorld.getAgent().getDirection() << endl;
      tmpWorld.doTurnNeckTo(lookDir);
    }

  bestStep = tmpWorld.getStep();
  tmpWorld.step();
  bestEval = evalDir(tmpWorld);

  // Second eval ballDir +-22.5, +-45
  TVector delta = destination - curWorld.getAgent().getPosition();
  TAngle turnDir = delta.getAngle() - curWorld.getAgent().getDirection();
  for(int i = -2; i <= 2; i++)
    {
      TAngle curDir = turnDir + i * 22.5;
      //      curDir = fmod(curDir + 720 + 180, 360) - 180;

      tmpWorld = curWorld;
      tmpWorld.doActualTurn(curDir);

      //      cout << "anglein: " << tmpWorld.getAgent().getDirection() << endl;

      if(lookAtBall)
	{
	  tmpWorld.doTurnNeckTowardsBall();
	}
      else
	{
	  leftLimit = tmpWorld.getAgent().getDirection() - 85;
	  rightLimit = tmpWorld.getAgent().getDirection() + 85;
	  lookDir = lookView->getAngle(f, 
				       AngleDeg((double)leftLimit), 
				       AngleDeg((double)rightLimit)).getDeg();
	  //    cout << "anglein: " << tmpWorld.getAgent().getDirection() << endl;
	  tmpWorld.doTurnNeckTo(lookDir);
	}

      curStep = tmpWorld.getStep();
      tmpWorld.step();

      curEval = evalDir(tmpWorld);

      if(curEval < bestEval)
	{
	  bestEval = curEval;
	  bestStep = curStep;
	}
    }

//   Display* display = Display::instance();
//   TVector tmp = worldStateView->getWorldState().getAgent().getDirectionVector();
//   tmp.rotate(turnDir);
//   display->addArrow(worldStateView->getWorldState().getAgent().getPosition(),
//  		    worldStateView->getWorldState().getAgent().getPosition()
//  		    + tmp * 5);

  Plan p = Plan(bestStep);
  return p;
}

/**
 * Persistent.
 * FALSE
 */
FuzzyBool MoveToPosition::persistent()
{
  return FALSE;
}

/**
 * Returns the applicable value
 */

FuzzyBool MoveToPosition::applicable()
{
  if(goalFound(worldStateView->getWorldState()))
    {
      return FALSE;
    }

  return TRUE;
}

RS_END_NAMESPACE

