/* -*- Mode: C++ -*-
 * $Id: PassableIntention.cc,v 1.17 2001/07/31 14:27:37 mikbr212 Exp $
 *
 * IDENTIFICATION
 *
 * $Revision: 1.17 $
 * Created by: Mikael Brnnstrm, Andreas Bckert, Andreas Sigfridsson
 * Last modified by $Author: mikbr212 $ $Date: 2001/07/31 14:27:37 $
 *
 *
 * HISTORY
 *
 * $Log: PassableIntention.cc,v $
 * Revision 1.17  2001/07/31 14:27:37  mikbr212
 * Changed threshold constant in getPosition
 *
 * Revision 1.16  2001/07/29 12:03:23  mikbr212
 * Now allows passable above offside line in some cases
 *
 * Revision 1.15  2001/07/29 11:41:11  sigge
 * Passable no longer considers possession but instead uses first interceptor.
 *
 * Revision 1.14  2001/07/29 09:43:30  mikbr212
 * Adjusted some things
 *
 * Revision 1.13  2001/07/28 16:08:34  mikbr212
 * Changed limits for far and very far
 *
 * Revision 1.12  2001/07/28 13:16:35  mikbr212
 * Changed looking
 *
 * Revision 1.11  2001/07/28 10:34:35  mikbr212
 * Fixed looking
 *
 * Revision 1.10  2001/07/27 13:30:44  mikbr212
 * Not much new
 *
 * Revision 1.9  2001/07/26 15:44:14  mikbr212
 * added offside avoidance and look at ball
 *
 * Revision 1.8  2001/07/26 12:19:33  mikbr212
 * Changed passtree
 *
 * Revision 1.7  2001/07/06 20:48:27  mikbr212
 * Added some reasonable utilities
 *
 * Revision 1.6  2001/07/06 14:47:40  andbo928
 * Patched to make it compile.
 *
 * Revision 1.5  2001/07/06 14:29:45  mikbr212
 * fixed utility
 *
 * Revision 1.4  2001/07/06 08:59:55  mikbr212
 * Made the some things work (but not yet in the new design)
 *
 * Revision 1.3  2001/07/03 15:24:19  andbo928
 * Added utility methods.
 *
 * Revision 1.2  2001/07/01 14:16:00  sigge
 * Added Mickes code (Intentions)
 *
 * Revision 1.1  2001/06/19 21:35:10  mikbr212
 * Created stubs
 *
 * Revision 1.1  2001/06/19 14:11:43  mikbr212
 * Created
 *
 */

/**
 * @file PassableIntention.cc
 * Implementation file for the class PassableIntention.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.17 $
 * $Author: mikbr212 $
 * $Date: 2001/07/31 14:27:37 $
 */


#include "PassableIntention.h"

#include "MoveToPosition.h"
#include "Display.h"

RS_BEGIN_NAMESPACE

/** Constructs an intention. 
 */
PassableIntention::PassableIntention()
  : Intention(),
    movements(),
    lookView(NULL)
{
  getView(lookView, LOOK_VIEW_ID);
  initMovements();
}

/** Destroys the intention. 
 */
PassableIntention::~PassableIntention() {
  viewManager->releaseView(LOOK_VIEW_ID);
}

/** Init movements evaluated in the pass tree.
 */
void PassableIntention::initMovements()
{
  const int innerDirs = 4;
  const float innerRadius = 2.5;
  for(int i=0; i<innerDirs; i++) 
    movements.push_back(TVector::TPolarVector(innerRadius, float(i)*360/float(innerDirs)));

  const int midDirs = 8;
  const float midRadius = 6.0;
  for(int i=0; i<midDirs; i++) 
    movements.push_back(TVector::TPolarVector(midRadius, float(i)*360/float(midDirs)));

  const int outerDirs = 6;
  const float outerRadius = 12.0;
  for(int i=0; i<outerDirs; i++) 
    movements.push_back(TVector::TPolarVector(outerRadius, float(i)*360/float(outerDirs)));
}

/** Get the best position (of the agent) from the pass tree.
 * Offside is avoided.
 */
TVector PassableIntention::getPosition() {
  const float fieldMargin(5.0);
  
  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());
  const TVector agentPosition(agent.getPosition());
  float offsideLine(worldStateView->getWorldState().getOffsideLine());
  if(ball.getPosition().getX() > offsideLine + 1)
    offsideLine = 52.5 - parameterView->SP_penalty_area_length;
  
  PassTree pt(passTreeView->getAgentPassTree());
  std::vector<TVector>::const_iterator it;
  float bestValue(0);
  TVector bestPosition;
  bool fallBack(false); 
  if(agentPosition.getX() > offsideLine)
    fallBack = true;

  float value;
  TVector position;
  for(it = movements.begin(); it != movements.end(); it++) {
    position = *it + agentPosition;
    if(position.getX() > offsideLine)
      continue;

    if(position.getY() < -(52.5 - fieldMargin) ||
       position.getY() > 52.5 - fieldMargin ||
       position.getX() < -(34.0 - fieldMargin) ||
       position.getX() > 34.0 - fieldMargin)
      continue;

    value = pt.evaluatePosition(position);
    if(value > bestValue) {
      bestValue = value;
      bestPosition = position;
    }
  }
  
  if(bestValue > 0.001) 
    return bestPosition;

  if(fallBack) // force fall back
    return TVector(offsideLine,
		   agentPosition.getY());
  

  return agentPosition; 
}

/** Checks wether the agent should look at the ball or a direction
 * choosen by LookView, while moving.
 */
bool PassableIntention::getLookAtBall() {
  PassTree::Analysis analysis(passTreeView->getAgentPassTree().getAnalysis());

  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());
  float distanceToBall((ball.getPosition() - agent.getPosition()).getLength());

  if((analysis.getDepthToRootPlayer() <= 2 && analysis.inScoreBranch()) 
     || distanceToBall < 10)
    return true;
  return false;
}

/* Generates a plan. 
 * Returns an empty plan.
 */
Plan PassableIntention::generatePlan() {
  static MoveToPosition moveSkill;
  // *** set position ***
  TVector position(getPosition());
  moveSkill.setPosition(position);

  // *** set look at ball ***
  moveSkill.setLookAtBall(getLookAtBall());

  // *** set speed ***
  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());
  
  float distance((ball.getPosition() - agent.getPosition()).getLength());
  const float farDist(15.0);
  const float veryFarDist(30.0);

  if(distance >= veryFarDist) // very far away from ball
    moveSkill.setSpeed(MoveToPosition::msGainStamina);
  else if(distance >= farDist) // far away from ball
    moveSkill.setSpeed(MoveToPosition::msKeepStamina);
  else { // not far away 
    float distToPos((agent.getPosition()-position).getLength());
    if(distToPos < 2.5)
      moveSkill.setSpeed(MoveToPosition::msKeepStamina);
    else
      moveSkill.setSpeed(MoveToPosition::msNormal);
  }

  if(moveSkill.applicable()) // ie, I am not standing there
    return moveSkill.generatePlan();
  
  // else turn towards the ball
  AgentState newAgent(agent);
  newAgent.doTurnTowards(ball.getPosition());
  if(getLookAtBall()) {
    // ... and look towards ball
    newAgent.doTurnNeckTowards(ball.getPosition());
  } else {
    // ... and look somewhere (from look view)
    float angleToBall = (ball.getPosition() - agent.getPosition()).getAngle();
    float delta(45); // FIXME: Check view width
    AngleDeg lookDir = lookView->getAngle(angleToBall - delta,
					  angleToBall + delta);
    newAgent.doTurnNeckTo(lookDir.getDeg());
  }
			
  return Plan(newAgent.getStep());
}

/** Models a player that is trying to get passable. 
 */
void PassableIntention::model(WorldState& ws) {
  // FIXME: Do something good here that doesn't 
  // require that much CPU... hmm, the player will stand still.
  ws.doNull();
}

/** Return @c the utility value of the intention.
 * The agent is considered.
 */
Intention::Utility PassableIntention::utility()
{
  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const BallState& ball(worldStateView->getWorldState().getBall());

  int ourInterceptionTime = statisticsView->getOurInterceptionTime();
  int theirInterceptionTime = statisticsView->getTheirInterceptionTime();

  if(ourInterceptionTime < theirInterceptionTime)
  {
    float distance((ball.getPosition() - agent.getPosition()).getLength());
    const float closeDist(15.0);
    const float medDist(30.0);
    const float farDist(40.0);
    if(distance < closeDist)
      return uVeryGood;
    else if(distance < medDist)
      return uGood;
    else if(distance < farDist)
      return uNormal;
    else
      return uBad;
  }
  else
    return uVeryBad;		// They have (or will have) the ball
}

/** Return @c the utility value of the intention.
 */
Intention::Utility PassableIntention::modelUtility(const WorldState& ws) {
  const AgentState& agent(ws.getAgent());
  const BallState& ball(ws.getBall());
  
  float distance((ball.getPosition() - agent.getPosition()).getLength());
  const float closeDist(8.0);
  const float farDist(15.0);

  if(distance >= farDist) // far away from ball
    return uNormal;
  else if(distance <= closeDist) // close to ball
    return uNormal;
  else // not close, not far away
    return uNormal;
}


/** Returns the persistence of this intention.
 * This intention is very persistent.
 * @return a float between zero (not persistent) and one (very persistent)
 */
float PassableIntention::getPersistence() const {
  return 1.0; // Very persistent
}

/** Returns the ID that identifies this (type of) intention.
 */
IntentionId PassableIntention::getID() const {
  return PASSABLE_INTENTION_ID;
}

RS_END_NAMESPACE

/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "PassableIntention.icc"
#endif







