/* -*- Mode: C++ -*-
 * $Id: FuzzyPlayer.cc,v 1.1.1.1 2001/04/08 11:33:59 sigge Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FuzzyPlayer.cc
 * Unitname: Library
 * $Revision: 1.1.1.1 $
 * Created by: Mikael Brnnstrm
 * Last modified by $Author: sigge $
 *
 *
 * HISTORY
 *
 * $Log: FuzzyPlayer.cc,v $
 * Revision 1.1.1.1  2001/04/08 11:33:59  sigge
 * Imported sources
 *
 */

/**
 * @file FuzzyPlayer.cc
 * Implementation of the class FuzzyPlayer.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.1.1.1 $
 * $Author: sigge $
 * $Date: 2001/04/08 11:33:59 $
 */


#include "FuzzyPlayer.h"

RS_BEGIN_NAMESPACE

const std::string WithUnknown<TeamBinary>::typeString = "TeamBinary";

FuzzyPlayer::FuzzyPlayer() 
  : pos(0,0),
    speed(0,0),
    uNum(0),
    neckDir(0),
    bodyDir(0),
    team(TB_OUR_TEAM)
{
  degrade(0);
}
			 

/*PointUCO _pos,
  VectorUCO _speed,
  TeamUCO _team,
  UInt16UCO _uNum,
  AngleDegUCO _bodyDir,
  AngleDegUCO _neckDir)
  : pos(_pos),
  speed(_speed),
  team(_team),
  uNum(_uNum),
  bodyDir(_bodyDir),
  neckDir(_neckDir) 
  {
  }
*/

Confidence FuzzyPlayer::vectorConfidence(const VectorUCO& vec) const {
  return vec.getAngle().getConfidence().groupWith(vec.getLength().getConfidence());
}

Confidence FuzzyPlayer::pointConfidence(const PointUCO& p) const {
  return p.getX().getConfidence().groupWith(p.getY().getConfidence());
}

FuzzyPlayer::FuzzyPlayer(const FuzzyPlayer& p1, const FuzzyPlayer& p2) 
  : team(TB_OUR_TEAM)
{
  if(pointConfidence(p1.pos) > pointConfidence(p2.pos))
    pos = p1.pos;
  else
    pos = p2.pos;

  if(vectorConfidence(p1.speed) > vectorConfidence(p2.speed))
    speed = p1.speed;
  else
    speed = p2.speed;

  if(p1.uNum.getConfidence() > p2.uNum.getConfidence())
    uNum = p1.uNum;
  else
    uNum = p2.uNum;

  if(p1.neckDir.getConfidence() > p2.neckDir.getConfidence())
    neckDir = p1.neckDir;
  else
    neckDir = p2.neckDir;

  if(p1.bodyDir.getConfidence() > p2.bodyDir.getConfidence())
    bodyDir = p1.bodyDir;
  else
    bodyDir = p2.bodyDir;

  if(p1.team.getConfidence() > p2.team.getConfidence())
    team = p1.team;
  else
    team = p2.team;
}

FuzzyPlayer::~FuzzyPlayer() {}

void FuzzyPlayer::fixUnknown() {
  FloatUCO f = FloatUCO(0);
  f.setConfidence(0.0);
  
  AngleDegUCO a = AngleDeg(0);
  a.setConfidence(0.0);

  TeamUCO t = TeamUCO(TB_OUR_TEAM);
  t.setConfidence(0.0);

  UInt16UCO i = UInt16UCO(0);
  i.setConfidence(0.0);

  if(!pos.getX().isKnown())
    pos.setX(f);
  if(!pos.getY().isKnown())
    pos.setY(f);

  if(!speed.getLength().isKnown())
    speed.setLength(f);
  if(!speed.getAngle().isKnown())
    speed.setAngle(a);

  if(!uNum.isKnown())
    uNum = i;

  if(!neckDir.isKnown())
    neckDir = a;

  if(!bodyDir.isKnown())
    bodyDir = a;

  if(!team.isKnown())
    team = t;
}


void FuzzyPlayer::degrade(float f) {
  FloatUCO x;
  AngleDegUCO a;

  // pos
  x = pos.getX();
  x.setConfidence(x.getConfidence() * f);
  pos.setX(x);
  x = pos.getY();
  x.setConfidence(x.getConfidence() * f);
  pos.setY(x);

  // speed
  a = speed.getAngle();
  a.setConfidence(a.getConfidence() * f);
  speed.setAngle(a);
  x = speed.getLength();
  x.setConfidence(x.getConfidence() * f);
  speed.setLength(x);
  
  // uNum
  uNum.setConfidence(uNum.getConfidence() * f);
  
  // neckDir
  neckDir.setConfidence(neckDir.getConfidence() * f);
  
  // bodyDir
  bodyDir.setConfidence(bodyDir.getConfidence() * f);
  
  // team
  team.setConfidence(team.getConfidence() * f);
}

void FuzzyPlayer::predict() {
  const float f = 0.9; // degrade factor
  degrade();

  speed = Vector(0,0);
  pos += speed.getVector(); 
  // degrade speed one extra time
  // team.setConfidence(team.getConfidence()*f);
}  

Confidence FuzzyPlayer::getConfidence() const {
  // Maybe combine all of them?
  return pointConfidence(pos);
}

FuzzyBool FuzzyPlayer::isMatch(const FuzzyPlayer& plr) {
  // UNum
  Float uNumF = (uNum == plr.uNum);
  adjustForConfidence(uNumF, plr.uNum.getConfidence(), uNum.getConfidence());
  
  // Team
  Float teamF = (team == plr.team);
  adjustForConfidence(teamF, plr.team.getConfidence(), team.getConfidence());
  
  // Position
  Float posF;
  Float posDist = 4.0;
  if(uNumF > 0.7 &&
     teamF > 0.7)
    posDist = 7.0;
  
  if(pos.isKnown() && plr.pos.isKnown()) {
    posF = 1.0 - VectorUCO(pos,plr.pos).getLength().getValue().getValue().getValue()/4.0;
    adjustForConfidence(posF,pointConfidence(pos),pointConfidence(plr.pos));
  } else
    posF = 0;
  
  // Speed
  Float speedF;
  
  if(speed.isKnown() && plr.speed.isKnown()) {
    speedF = 1.0 - (speed - plr.speed).getLength().getValue().getValue().getValue()/0.9;
    adjustForConfidence(speedF,vectorConfidence(speed),vectorConfidence(plr.speed));
  } else
    posF = 0;
  
   // don't care about bodyDir and neckDir

  return FuzzyBool(posF * (2.0 + 2.0*uNumF + 2.0*teamF + 1.0*speedF)/7.0);
} 

void FuzzyPlayer::adjustForConfidence(Float& f,
				      const Confidence& c1, 
				      const Confidence& c2) const {
  Confidence c(c1);
  f = minMax(0.0, f, 1.0);
  f = (f - 0.5) * c.combineWith(c2).getConfidence() + 0.5;
}

/*FuzzyPlayer FuzzyPlayer::union(const FuzzyPlayer& plr) {
  return FuzzyPlayer(*this, plr); 
  }*/

RS_END_NAMESPACE


#include "FuzzyPlayer.icc"



